#ifndef I386_ATOMIC_H
#define I386_ATOMIC_H

static inline int cmpxchg32(volatile u32 *m, u32 old, u32 new)
{
	__asm__ __volatile__("lock; cmpxchgl %1,%2"
			     : "=a"(old)
			     : "r"(new), "m"(*m), "0"(old)
			     : "memory");
	return old;
}

#define cmpxchgptr(m, old, new) cmpxchg32((volatile u32*) (m), (u32) (old), (u32) (new))

#define xchgptr(new, mem) __xchgptr((void*) new, (void**) mem)

static __inline__ u32 xadd32(u32 i, volatile u32* mem)
{
	u32 inc = i;
	__asm__ __volatile__(
		"lock; xaddl %0, %1"
		:"+r" (i), "+m" (*mem)
		: : "memory");
	return i + inc;
}

static __inline__ void* __xchgptr(void* new, void** mem)
{
	__asm__ __volatile__("xchgl %0,%1"
			     :"=r" (new)
			     :"m" (*mem), "0" (new)
			     :"memory");
	return new;
}


typedef struct { volatile int counter; } atomic_t;

#define ATOMIC_INIT(i)	{ (i) }

/**
 * atomic_read - read atomic variable
 * @v: pointer of type atomic_t
 *
 * Atomically reads the value of @v.
 */
#define atomic_read(v)		((v)->counter)

/**
 * atomic_set - set atomic variable
 * @v: pointer of type atomic_t
 * @i: required value
 *
 * Atomically sets the value of @v to @i.
 */
#define atomic_set(v,i)		(((v)->counter) = (i))

/**
 * atomic_add_return - add and return
 * @v: pointer of type atomic_t
 * @i: integer value to add
 *
 * Atomically adds @i to @v and returns @i + @v
 */
static __inline__ int atomic_add_return(int i, atomic_t *v)
{
	return (int) xadd32((u32) i, (u32*)&v->counter);
}

static __inline__ int atomic_add_return_old(int i, atomic_t *v)
{
	__asm__ __volatile__(
		"lock; xaddl %0, %1"
		:"+r" (i), "+m" (v->counter)
		: : "memory");
	return i;
}


static __inline__ void atomic_add(int i, atomic_t *v)
{
	atomic_add_return(i, v);
}

#define atomic_inc_return(v)  (atomic_add_return(1,v))
#define atomic_inc(v) (atomic_add(1, v))

#endif
