#ifndef LINRW_H
#define LINRW_H

/* from Linux 2.6.28 */

typedef struct {
	unsigned int lock;
} raw_rwlock_t;

#define RW_LOCK_BIAS		 0x01000000

#define __RAW_RW_LOCK_UNLOCKED		{ RW_LOCK_BIAS }

# define LOCK_PTR_REG "a"
# define REG_PTR_MODE "k"

#define LOCK_PREFIX "lock; "

static inline void __raw_read_lock(raw_rwlock_t *rw)
{
	asm volatile(LOCK_PREFIX " subl $1,(%0)\n\t"
		     "jns 1f\n"
		     "call __read_lock_failed\n\t"
		     "1:\n"
		     ::LOCK_PTR_REG (rw) : "memory");
}

static inline void __raw_write_lock(raw_rwlock_t *rw)
{
	asm volatile(LOCK_PREFIX " subl %1,(%0)\n\t"
		     "jz 1f\n"
		     "call __write_lock_failed\n\t"
		     "1:\n"
		     ::LOCK_PTR_REG (rw), "i" (RW_LOCK_BIAS) : "memory");
}

static inline void __raw_read_unlock(raw_rwlock_t *rw)
{
	asm volatile(LOCK_PREFIX "incl %0" :"+m" (rw->lock) : : "memory");
}

static inline void __raw_write_unlock(raw_rwlock_t *rw)
{
	asm volatile(LOCK_PREFIX "addl %1, %0"
		     : "+m" (rw->lock) : "i" (RW_LOCK_BIAS) : "memory");
}


#define lin_read_lock(p)	__raw_read_lock(p)
#define lin_read_unlock(p)	__raw_read_unlock(p)
#define lin_write_lock(p)	__raw_write_lock(p)
#define lin_write_unlock(p)	__raw_write_unlock(p)

typedef raw_rwlock_t lin_rwlock_t;

#define DEFINE_LIN_RWLOCK(x) lin_rwlock_t x = { RW_LOCK_BIAS }


#endif
