#ifndef SPARC64_ATOMIC_H
#define SPARC64_ATOMIC_H

static inline int
cmpxchg32(volatile u32 *m, u32 old, u32 new)
{
	__asm__ __volatile__("membar #StoreLoad | #LoadLoad\n"
			     "cas [%2], %3, %0\n\t"
			     "membar #StoreLoad | #StoreStore"
			     : "=&r" (new)
			     : "0" (new), "r" (m), "r" (old)
			     : "memory");

	return new;
}

static inline long
cmpxchg64(volatile u64 *m, u64 old, u64 new)
{
	__asm__ __volatile__("membar #StoreLoad | #LoadLoad\n"
			     "casx [%2], %3, %0\n\t"
			     "membar #StoreLoad | #StoreStore"
			     : "=&r" (new)
			     : "0" (new), "r" (m), "r" (old)
			     : "memory");
	return new;
}

static __inline__ u32 xadd32(u32 i, volatile u32* mem)
{
	u32 old;
	u32 ret;
	goto first;
	do {
		cpu_relax();
	first:
		old = *mem;
		ret = cmpxchg32(mem, old, old + i);
	} while (ret != old);
	return (old + i);
}

#define cmpxchgptr(m, old, new) cmpxchg64((volatile u64*) (m), (u64) (old), (u64) (new))

#define xchgptr(new, mem) __xchgptr((void*) new, (void**) mem)

static __inline__ void* __xchgptr(void* new, void** mem)
{
	void* old;
	void* ret;
	goto first;
	do {
		cpu_relax();
	first:
		old = (void*) *mem;
		ret = (void*) cmpxchgptr(mem, old, new);
	} while (ret != old);
	return old;
}


typedef struct { volatile int counter; } atomic_t;

#define ATOMIC_INIT(i)	{ (i) }

/**
 * atomic_read - read atomic variable
 * @v: pointer of type atomic_t
 *
 * Atomically reads the value of @v.
 */
#define atomic_read(v)		((v)->counter)

/**
 * atomic_set - set atomic variable
 * @v: pointer of type atomic_t
 * @i: required value
 *
 * Atomically sets the value of @v to @i.
 */
#define atomic_set(v,i)		(((v)->counter) = (i))

/**
 * atomic_add_return - add and return
 * @v: pointer of type atomic_t
 * @i: integer value to add
 *
 * Atomically adds @i to @v and returns @i + @v
 */
static __inline__ int atomic_add_return(int i, atomic_t *v)
{
	u32 old;
	u32 ret;
	goto first;
	do {
		cpu_relax();
	first:
		old = atomic_read(v);
		ret = cmpxchg32((u32*) &v->counter, old, old + (u32) i);
	} while (ret != old);
	return (int) (old + (u32) i);
}

static __inline__ int atomic_add_return_old(int i, atomic_t *v)
{
	u32 old;
	u32 ret;
	goto first;
	do {
		cpu_relax();
	first:
		old = atomic_read(v);
		ret = cmpxchg32((u32*) &v->counter, old, old + (u32) i);
	} while (ret != old);
	return (int) old;
}


static __inline__ void atomic_add(int i, atomic_t *v)
{
	atomic_add_return(i, v);
}

#define atomic_inc_return(v)  (atomic_add_return(1,v))
#define atomic_inc(v) (atomic_add(1, v))

#endif
