#ifndef __MCS_H__
#define __MCS_H__

#include "mem.h"
#include "atomic.h"

typedef struct __mcsnode
{
	int	           blocked;
	struct __mcsnode* next;	
} mcsnode_t;

typedef struct
{
	mcsnode_t*	tail;
} mcslock_t;

#define DEFINE_MCSLOCK(name) mcslock_t name = { NULL }
#define DEFINE_MCSNODE(name) mcsnode_t name

static inline void mcs_lock(mcslock_t* lock, mcsnode_t *self)
{
	mcsnode_t* prev;
	unsigned long tmp;

	self->next     = NULL;

	__asm__ __volatile__(
		"1:	ldx	[%2], %0	\n" /* get prev */
		"	mov	%3, %1		\n"
		"	casx	[%2], %0, %1	\n" /* enqueue self */
		"	membar	#StoreLoad | #StoreStore \n"
		"	cmp	%0, %1		\n"
		"	bne,pn	%%icc, 1b	\n"
		"	  cmp	%0, 0		\n" /* are we blocked? */
		"	bne,pn	%%icc, 2f	\n"
		"       .subsection 2		\n" /* => spin */
		"2:	mov	1, %1		\n"
		"	stuw	%1, [%3]	\n" /* set blocked */
		"	membar	#StoreStore	\n"
		"	stx	%3, [%0+8]	\n" /* set prev->next */
		"3:	lduw	[%3], %1	\n" /* load blocked */
		"	cmp	%1, 0		\n" /* are we unblocked? */
		"	bne,pt	%%icc, 3b	\n"
		"	  membar #LoadLoad	\n"
		"	ba,a,pt	%%icc, 4f	\n"
		"	.previous		\n"
		"4:	"
		: "=&r" (prev), "=&r" (tmp)
		: "r" (lock), "r" (self)
		: "memory");
}

static inline void mcs_unlock(mcslock_t* lock, mcsnode_t *self)
{
	unsigned long tmp;

	__asm__ __volatile__(
		"	ldx	[%2+8], %0	\n"
		"	cmp	%0, 0		\n"
		"	bne,pt	%%icc, 2f	\n" /* next already set? */
		"	  membar #LoadStore | #StoreStore \n"
		"	mov	%%g0, %0	\n"
		"	casx	[%1], %2, %0	\n" /* expect self   */
		"	cmp	%2, %0		\n"
		"	be,pt	%%icc, 3f	\n" /* did a newcomer race? */
		"1:	  ldx	[%2+8], %0	\n"
		"	cmp	%0, 0		\n"
		"	be,pn	%%icc, 1b	\n" /* wait for next != 0 */
		"	  membar #LoadLoad	\n"
		"2:	stuw	%%g0, [%0]	\n" /*self->next->blocked = 0*/
		"3:	"
		: "=&r" (tmp)
		: "r" (lock), "r" (self)
		: "memory");
}


#endif

