#ifndef MCS_RW_QLOCK
#define MCS_RW_QLOCK

#include "mem.h"
#include "atomic.h"
#include "locks/mcsrw_common.h"
#include "locks/mcsrw_proto.h"

/*
 *  Offsets mcs_rwlock_t:
 *          0 ->count
 *          8 ->tail
 *         16 ->nextw
 *  Offsets mcs_rwnode_t:
 *          0 ->cls
 *          8 ->next
 *         16 ->u
 *         16 ->u.state.blocked
 *         18 ->u.state.next_cls
 */

static inline void mcs_write_lock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	mcs_rwnode_t* prev;
	unsigned long tmp, tmp2;

	*self = (mcs_rwnode_t) {.cls = WRITER, .next = NULL,
				.u.state.blocked = 1, .u.state.next_cls = NONE};

	__asm__ __volatile__(
		"	membar	#StoreLoad | #StoreStore	\n"
		"	add	%3, 8, %2			\n"
		"1:	ldx	[%2], %0			\n" /*tail*/
		"	mov	%4, %1				\n" /*self*/
		"	casx	[%2], %0, %1			\n" /*tail*/
		"	cmp	%0, %1				\n"
		"	bne,pn	%%icc, 1b			\n"
		"	  cmp	%0, 0				\n"
		"	bne,a,pn %%icc, 2f			\n"
		"	  mov	2, %1				\n"
		"	stx	%4, [%3+16]			\n" /*nextw*/
		"	membar	#StoreLoad			\n"
		"	lduw	[%3], %1			\n" /*count*/
		"	cmp	%1, 0				\n" /*reader?*/
		"	bne,pn	%%icc, 3f			\n" /*spin*/
		"	  clr	%1				\n"
		"	add	%3, 16, %2			\n"
		"	casx	[%2], %4, %1			\n" /*nextw*/
		"	cmp	%4, %1				\n"
		"	bne,a,pn %%icc, 5f			\n"
		"	  ldx	[%2], %0			\n" /*nextw*/
		"4:						\n"
		"	.subsection 2				\n"
		"5:	casx	[%2], %0, %1			\n" /*nextw*/
		"	cmp	%0, %1				\n"
		"	bne,a,pn %%icc, 5b			\n"
		"	  ldx	[%2], %0			\n" /*nextw*/
		"	ba,a	%%icc, 3f			\n" /*spin*/
		"2:	stuh	%1, [%0+18]			\n" /*nxt_cls*/
		"	membar	#StoreStore			\n"
		"	stx	%4, [%0+8]			\n" /*next*/
		"3:	lduh	[%4+16], %1			\n" /*blocked*/
		"	cmp	%1, 0				\n"
		"	bne,a,pn %%icc, 3b			\n"
		"	  membar #LoadLoad			\n"
		"	ba,a	%%icc, 4b			\n"
		"	.previous"
		: "=&r" (prev), /* %0 */
		  "=&r" (tmp),  /* %1 */
		  "=&r" (tmp2)  /* %2 */
		: "r" (lock),   /* %3 */
		  "r" (self)    /* %4 */
		: "memory"
		);
}


static inline void mcs_write_unlock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	unsigned long tmp, tmp2;

	__asm__ __volatile__(
		"	membar	#LoadStore | #StoreStore	\n"
		"	ldx	[%2+8], %0			\n" /*next*/
		"	cmp	%0, 0				\n"
		"	bne,a,pn %%icc, 2f			\n"
		"	  lduw	[%0], %1			\n" /*nextcls*/
		"	add	%3, 8, %1			\n"
		"	clr	%0				\n"
		"	casx	[%1], %2, %0			\n" /*tail*/
		"	cmp	%0, %2				\n"
		"	bne,a,pn %%icc, 1f			\n"
		"	  ldx	[%2+8], %0			\n" /*next*/
		"5:						\n"
		"	.subsection 2				\n"
		/*      spin loop: await self->next != 0	 */
		"1:	membar	#LoadLoad			\n"
		"	cmp	%0, 0				\n"
		"	be,a,pn	%%icc, 1b			\n"
		"	  ldx	[%2+8], %0			\n" /*next*/
		"	lduw	[%0], %1			\n" /*nextcls*/
		/*	inc count if next is reader		 */
		"2:	cmp	%1, 1				\n"
		"	bne,a,pn %%icc, 4f			\n"
		"	  nop					\n"
		"3:	lduw	[%3], %1			\n" /*count*/
		"	add	%1, 1, %2			\n"
		"	cas	[%3], %1, %2			\n"
		"       cmp	%1, %2				\n"
		"	bne,a,pn %%icc, 3b			\n"
		"	  membar #LoadLoad			\n"
		/*	unblock next				 */
		"4:	stuh	%%g0, [%0+16]			\n"
		"	ba,a	%%icc, 5b			\n"
		"	.previous				\n"
		: "=&r" (tmp),  /* %0 */
		  "=&r" (tmp2), /* %1 */
		  "+r" (self)   /* %2 */
		: "r" (lock)    /* %3 */
		: "memory"
		);
}

static inline void mcs_read_lock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	unsigned long tmp, tmp2;
	mcs_rwnode_t* prev;

	*self = (mcs_rwnode_t) {.cls = READER, .next = NULL,
				.u.state.blocked = 1, .u.state.next_cls = NONE};

	__asm__ __volatile__(
		"	membar	#StoreLoad | #StoreStore	\n"
		"	add	%3, 8, %2			\n"
		"1:	ldx	[%2], %0			\n"
		"	mov	%4, %1				\n"
		"	casx	[%2], %0, %1			\n" /*tail*/
		"	cmp	%0, %1				\n"
		"	bne,pn	%%icc, 1b			\n"
		"	  cmp	%0, 0				\n"
		"	bne,a,pn %%icc, 2f			\n"
		"	  lduw	[%0], %1			\n" /*prevcls*/
		/*	!prev					 */
		"3:	lduw	[%3], %0			\n" /*count*/
		"	add	%0, 1, %1			\n"
		"	cas	[%3], %0, %1			\n"
		"	cmp	%1, %0				\n"
		"	bne,a,pn %%icc, 3b			\n"
		"	  membar #LoadLoad			\n"
		"	stuh	%%g0, [%4+16]			\n" /*blocked*/
		"9:	membar	#StoreLoad			\n"
		/*	check for next reader			 */
		"	lduh	[%4+18], %0			\n" /*nextcls*/
		"	cmp	%0, 1				\n"
		"	be,a,pt	%%icc, 4f			\n"
		"	  ldx	[%4+8], %2			\n" /*next*/
		"6:						\n"

		"	.subsection 2				\n"
		/*	next_cls == READER			 */
		"4:	cmp	%2, 0				\n"
		"	membar	#LoadLoad			\n"
		"	be,a,pn	%%icc, 4b			\n"
		"	  ldx	[%4+8], %2			\n" /*next*/
		"5:	lduw	[%3], %0			\n" /*count*/
		"	add	%0, 1, %1			\n"
		"	cas	[%3], %0, %1			\n"
		"	cmp	%1, %0				\n"
		"	bne,a,pn %%icc, 5b			\n"
		"	  membar #LoadLoad			\n"
		"	membar	#StoreStore			\n"
		"	stuh	%%g0, [%2+16]			\n" /*blocked*/
		"	ba,a	%%icc, 6b			\n"
		/*	prev != 0				 */
		"2:	cmp	%1, 2				\n"
		"	be,pt %%icc, 7f				\n" /*writer?*/
		"	  sethi	%%hi(0x10000), %1		\n"
		"	add	%0, 16, %0			\n"
		"	or	%1, 1, %2			\n"
		"	cas	[%0], %1, %2			\n" /*prev.u*/
		"	cmp	%1, %2				\n"
		"	bne,pn	%%icc, 11f			\n"
		"	  sub	%0, 16, %0			\n"
		"	membar	#StoreStore			\n"
		"7:	stx	%4, [%0+8]			\n" /*prevnxt*/
		"	membar	#StoreLoad			\n"
		"8:	lduh	[%4+16], %1			\n" /*blocked*/
		"	cmp	%1, 0				\n"
		"	bne,a,pn %%icc, 8b			\n"
		"	  membar #LoadLoad			\n"
		"	ba,a	%%icc, 9b			\n"
		/*	not a writer and cas failed		 */
		"11:	lduw	[%3], %2			\n" /*count*/
		"	add	%2, 1, %1			\n"
		"	cas	[%3], %2, %1			\n"
		"	cmp	%1, %2				\n"
		"	bne,a,pn %%icc, 11b			\n"
		"	  membar #LoadLoad			\n"
		"	membar	#StoreStore			\n"
		"	stx	%4, [%0+8]			\n" /*prevnxt*/
		"	stuh	%%g0, [%4+16]			\n" /*blocked*/
		"	ba,a	%%icc, 9b			\n"
		"	.previous				\n"
		: "=&r" (prev), /* %0 */
		  "=&r" (tmp),  /* %1 */
		  "=&r" (tmp2)  /* %2 */
		: "r" (lock),   /* %3 */
		  "r" (self)    /* %4 */
		: "memory"
		);
}

static inline void mcs_read_unlock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	mcsp_read_unlock(lock, self);
}



#endif
