#ifndef MCSSRW_H
#define MCSSRW_H

#include "mem.h"

typedef struct {
	u32 in, out;
} mcss_rwlock_t;

#define RC_INCR   0x10000
#define WC_INCR   0x1
#define W_MASK    0xffff

#define WC_TOPMSK 0x8000
#define RC_TOPMSK 0x80000000

#define DEFINE_MCSS_RWLOCK(x) mcss_rwlock_t x =	\
	{ 0 }

static inline void mcss_read_lock(mcss_rwlock_t *l)
{
	unsigned long tmp1, tmp2, tmp3;

	__asm__ __volatile__(
"1:	lduw		[%3], %0	\n"
"	sethi     	%%hi(0x10000), %2 \n"
"	add		%0, %2, %1	\n"
"	sethi		%%hi(0x7fffffff), %2 \n"
"	or		%%lo(0x7fffffff), %2, %2 \n"
"	and		%1, %2, %1 \n"
"	cas		[%3], %0, %1	\n"
"	cmp		%0, %1		\n"
"	bne,pn       	%%icc, 1b	\n"
"	 membar		#StoreLoad | #StoreStore \n"
"	sethi		%%hi(0xffff), %2 \n"
"	or		%%lo(0xffff), %2, %2 \n"
"	and		%1, %2, %1	\n"
"2:	lduw		[%4], %0	\n"
"	and		%0, %2, %0	\n"
"	cmp		%0, %1		\n"
"	bne,pn	%%icc, 2b		\n"
"	 membar		#LoadLoad	\n"
: "=&r" (tmp1), "=&r" (tmp2), "=&r" (tmp3)
: "r" (&l->in), "r" (&l->out)
	: "memory");
}

static inline void mcss_read_unlock(mcss_rwlock_t *l)
{
	unsigned long tmp1, tmp2, tmp3;

	__asm__ __volatile__(
"	membar		#StoreLoad | #LoadLoad \n"
"1:	lduw		[%3], %0	\n"
"	sethi     	%%hi(0x10000), %2 \n"
"	add		%0, %2, %1	\n"
"	sethi		%%hi(0x7fffffff), %2 \n"
"	or		%%lo(0x7fffffff), %2, %2 \n"
"	and		%1, %2, %1 \n"
"	cas		[%3], %0, %1	\n"
"	cmp		%0, %1		\n"
"	bne,pn       	%%icc, 1b	\n"
"	 nop				\n"
: "=&r" (tmp1), "=&r" (tmp2), "=&r" (tmp3)
	: "r" (&l->out)
	: "memory");
}


static inline void mcss_write_lock(mcss_rwlock_t *l)
{
	unsigned long tmp1, tmp2, tmp3;

	__asm__ __volatile__(
"	sethi		%%hi(0xffff7fff), %2 \n"
"       or		%%lo(0xffff7fff), %2, %2 \n"
"1:	lduw		[%3], %0	\n"
"	add		%0, 0x1, %1	\n"
"	and		%1, %2, %1\n"
"	cas		[%3], %0, %1	\n"
"	cmp		%0, %1		\n"
"	bne,pn       	%%icc, 1b	\n"
"	 membar		#StoreLoad | #StoreStore \n"
"2:	lduw		[%4], %0	\n"
"	cmp		%0, %1		\n"
"	bne,pn	%%icc, 2b		\n"
"	 membar		#LoadLoad	\n"
: "=&r" (tmp1), "=&r" (tmp2), "=&r" (tmp3)
        : "r" (&l->in), "r" (&l->out)
	: "memory");
}

static inline void mcss_write_unlock(mcss_rwlock_t *l)
{
	unsigned long tmp1, tmp2, tmp3;;

	__asm__ __volatile__(
"	sethi		%%hi(0xffff7fff), %2 \n"
"       or		%%lo(0xffff7fff), %2, %2 \n"
"	membar		#StoreLoad | #LoadLoad \n"
"1:	lduw		[%3], %0	\n"
"	add		%0, 0x1, %1	\n"
"	and		%1, %2, %1\n"
"	cas		[%3], %0, %1	\n"
"	cmp		%0, %1		\n"
"	bne,pn       	%%icc, 1b	\n"
"	 nop				\n"
: "=&r" (tmp1), "=&r" (tmp2), "=&r" (tmp3)
	: "r" (&l->out)
	: "memory");
}

#endif
