#ifndef TICKETLOCK_H
#define TICKETLOCK_H

typedef struct {
	int tail;
	int head;
} raw_ticketlock_t;

typedef raw_ticketlock_t ticketlock_t;

#define DEFINE_TICKETLOCK(x) ticketlock_t x =\
	{ 0, 0 }

#define __raw_ticket_is_locked(lp)	((lp)->tail != (lp)->head)

#define __raw_ticket_unlock_wait(lp)	\
	do {	rmb();			\
	} while((lp)->tail != (lp)->head)



static inline void __raw_ticket_lockX(raw_ticketlock_t *lock)
{
	int ticket, tmp, tmp2;
	__asm__ __volatile__(
"1:	lduw		[%3], %0 \n" /* read ticket */
"	lduw		[%4], %2 \n"
"	add		%0, 1, %1 \n"
"	cas		[%3], %0, %1 \n"
"	cmp		%0, %1 \n"
"	bne,a,pn	%%icc, 1b \n"
"	 nop \n"
"2: \n"
"	cmp		%0, %2 \n"
"	be,pt		%%icc, 3f \n"
"	 nop \n"
"	lduw		[%4], %2 \n"
//"	udiv		%%o1, %0, %1\n"
//"	udiv		%%o1, %0, %1\n"
//" 	nop\n "
"	ba		2b\n"
"3:	membar		#StoreStore | #StoreLoad"
	: "=&r" (ticket), "=&r" (tmp), "=&r" (tmp2)
	: "r" (&lock->tail), "r" (&lock->head)
	: "memory");
}



static inline void __raw_ticket_lockG(raw_ticketlock_t *lock)
{
	int ticket, tmp;
	__asm__ __volatile__(
"1:     lduw		[%2], %0 \n" /* read ticket */
"	add		%0, 1, %1 \n"
"	cas		[%2], %0, %1 \n"
"	cmp		%0, %1 \n"
"	bne,a,pn	%%icc, 1b \n"
"	 nop \n"
"2:	lduw		[%3], %1 \n"
"	cmp		%0, %1 \n"
"	be,a,pt		%%icc, 3f \n"
"	 nop \n"
"	membar		#LoadLoad | #StoreLoad | #LoadStore\n"
//"	udiv		%%o1, %0, %1\n"
//"	udiv		%%o1, %0, %1\n"
//" 	nop\n "
"	ba		2b\n"
"3:	membar		#StoreStore | #StoreLoad"
	: "=&r" (ticket), "=&r" (tmp)
	: "r" (&lock->tail), "r" (&lock->head)
	: "memory");
}

static inline void __raw_ticket_lock(raw_ticketlock_t *lock)
{
	int ticket, tmp;
	__asm__ __volatile__(
"1:     lduw		[%2], %0 \n" /* read ticket */
"	add		%0, 1, %1 \n"
"	cas		[%2], %0, %1 \n"
"	cmp		%0, %1 \n"
"	be,a,pt		%%icc, 2f \n"
"	 nop \n"
"	membar		#LoadLoad | #StoreLoad | #LoadStore\n"
"	ba		1b\n"
"	 nop \n"
"2:	lduw		[%3], %1 \n"
"	cmp		%0, %1 \n"
"	be,a,pt		%%icc, 3f \n"
"	 nop \n"
"	membar		#LoadLoad | #StoreLoad | #LoadStore\n"
//"	udiv		%%o1, %0, %1\n"
//"	udiv		%%o1, %0, %1\n"
//" 	nop\n "
"	ba		2b\n"
"3:	membar		#StoreStore | #StoreLoad"
	: "=&r" (ticket), "=&r" (tmp)
	: "r" (&lock->tail), "r" (&lock->head)
	: "memory");
}

static inline void __raw_ticket_lockK(raw_ticketlock_t *lock)
{
        int ticket, tmp;
        __asm__ __volatile__(
		"1:     lduw            [%2], %0 \n" /* read ticket */
"       add             %0, 1, %1 \n"
"       cas             [%2], %0, %1 \n"
"       cmp             %0, %1 \n"
"       bne,a,pn        %%icc, 1b \n"
"        nop \n"
"2:     lduw            [%3], %1 \n"
"       cmp             %0, %1 \n"
"       bne,a,pn        %%icc, 2b \n"
"        nop \n"
"       membar          #StoreStore | #StoreLoad"
		: "=&r" (ticket), "=&r" (tmp)
		: "r" (&lock->tail), "r" (&lock->head)
		: "memory");
}


static inline int __raw_ticket_trylock(raw_ticketlock_t *lock)
{
	int tail, head;
	__asm__ __volatile__(
"	lduw		[%2], %0 \n" /* read tail */
"	lduw		[%3], %1 \n" /* read head */
"	cmp		%0, %1 \n"
"	bne,a,pn	%%icc, 1f \n"
"	 nop \n"
"	inc		%1 \n"
"	cas		[%2], %0, %1 \n" /* try to inc ticket */
"	membar		#StoreStore | #StoreLoad \n"
"1:	"
	: "=&r" (tail), "=&r" (head)
	: "r" (&lock->tail), "r" (&lock->head)
	: "memory");

	return tail == head;
}

static inline void __raw_ticket_unlock(raw_ticketlock_t *lock)
{
	int tmp;
	__asm__ __volatile__(
"	membar		#StoreStore | #LoadStore \n"
"	lduw		[%1], %0 \n"
"	inc		%0 \n"
"	st		%0, [%1] \n"
"	membar		#StoreStore | #StoreLoad"
	: "=&r" (tmp)
	: "r" (&lock->head)
	: "memory");
}

/* We don't handle this yet, but it looks like not re-enabling the interrupts
 * works fine, too. For example, lockdep also does it like this.
 */
#define __raw_ticket_lock_flages(l, f) __raw_ticket_lock(l)


#define ticket_lock(x)    __raw_ticket_lock(x)
#define ticket_unlock(x)  __raw_ticket_unlock(x)
#define ticket_trylock(x) __raw_ticket_trylock(x)

#endif 
