#ifndef BARW_PROTO2_H
#define BARW_PROTO2_H

#include "mem.h"

typedef struct {
	volatile u32 win, rin;
	volatile u32 wout, rout;
} ba_proto2_rwlock_t;

#define DEFINE_BA_PROTO2_RWLOCK(x) ba_proto2_rwlock_t x =	\
	{ 0, 0, 0, 0 }

#define BAP2_READER_BIAS (1 << 8)
#define BAP2_READER_MASK (0xff)

static inline void ba_proto2_read_lock(ba_proto2_rwlock_t *l)
{
	u32 blocked = xadd32(BAP2_READER_BIAS, &l->rin) & BAP2_READER_MASK;
	while (blocked && ((l->rin & BAP2_READER_MASK) == blocked))
		cpu_relax();
}

static inline void ba_proto2_read_unlock(ba_proto2_rwlock_t *l)
{
	xadd32(BAP2_READER_BIAS, &l->rout);
}

static inline void ba_proto2_write_lock(ba_proto2_rwlock_t *l)
{
	u32 ticket;
	ticket = xadd32(1, &l->win) - 1;
	while (ticket != l->wout)  {
		cpu_relax();
	}
	ticket = xadd32(0x2 | (ticket & 0x1), &l->rin) & (~(u32)0x3);
	while (ticket != l->rout)
		cpu_relax();
}

static inline void ba_proto2_write_unlock(ba_proto2_rwlock_t *l)
{
	*lsb32((u32*) &l->rin) = 0;
	mb();
	l->wout++;
}

#endif
