#ifndef LOCKS_H
#define LOCKS_H

#include "kernel_compat.h"
#include "mem.h"
#include "atomic.h"

#if defined(USE_SPINLOCK)

#include "locks/spin.h"
#define DEF_LOCK(x) DEFINE_SPINLOCK(x)
#define DEF_LOCK_CTX(x)
#define lock(x, ctx)     spin_lock(x)
#define unlock(x, ctx)   spin_unlock(x)

#elif defined(USE_PROTO1LOCK)

#include "locks/ticket_proto1.h"
#define DEF_LOCK(x) DEFINE_PROTO1LOCK(x)
#define DEF_LOCK_CTX(x)
#define lock(x, ctx)     ticket_proto1_lock(x)
#define unlock(x, ctx)   ticket_proto1_unlock(x)

#elif defined(USE_PROTO2LOCK)

#include "locks/ticket_proto2.h"
#define DEF_LOCK(x) DEFINE_PROTO2LOCK(x)
#define DEF_LOCK_CTX(x)
#define lock(x, ctx)     ticket_proto2_lock(x)
#define unlock(x, ctx)   ticket_proto2_unlock(x)

#elif defined(USE_TICKETLOCK)

#include "locks/ticket.h"
#define DEF_LOCK_CTX(x)
#define DEF_LOCK(x) DEFINE_TICKETLOCK(x)
#define lock(x, ctx)     ticket_lock(x)
#define unlock(x, ctx)   ticket_unlock(x)

#elif defined(USE_PROTOMCSLOCK)

#include "locks/mcs_proto.h"
#define DEF_LOCK_CTX(x) mcspnode_t x;
#define DEF_LOCK(x) DEFINE_MCSPLOCK(x)
#define HAVE_LOCK_CTX
#define lock(x, ctx)     mcsp_lock(x, &(ctx))
#define unlock(x, ctx)   mcsp_unlock(x, &(ctx))

#elif defined(USE_MCSLOCK)

#include "locks/mcs.h"
#define DEF_LOCK_CTX(x) mcsnode_t x;
#define DEF_LOCK(x) DEFINE_MCSLOCK(x)
#define HAVE_LOCK_CTX
#define lock(x, ctx)     mcs_lock(x, &(ctx))
#define unlock(x, ctx)   mcs_unlock(x, &(ctx))

#elif  defined(USE_LINRW_LOCK)

#include "locks/linrw.h"
#define DEF_LOCK(x) DEFINE_LIN_RWLOCK(x)
#define DEF_LOCK_CTX(x)
#define lock(x, ctx)     lin_write_lock(x)
#define unlock(x, ctx)   lin_write_unlock(x)


#elif defined(USE_MCSRW_LOCK)

#include "locks/mcsrw.h"
#define DEF_LOCK(x) DEFINE_MCS_RWLOCK(x)
#define DEF_LOCK_CTX(x) DEFINE_MCS_RWNODE(x)
#define HAVE_LOCK_CTX
#define lock(x, ctx)     mcs_write_lock(x, &ctx)
#define unlock(x, ctx)   mcs_write_unlock(x, &ctx)

#elif defined(USE_BARW_PROTO1_LOCK)

#include "locks/barw_proto1.h"
#define DEF_LOCK(x) DEFINE_BA_PROTO1_RWLOCK(x)
#define DEF_LOCK_CTX(x)
#define lock(x, ctx)     ba_proto1_write_lock(x)
#define unlock(x, ctx)   ba_proto1_write_unlock(x)

#else

#error "What kind of lock should I use?"

#endif


#endif
