#ifndef __Q_LOCK_H__
#define __Q_LOCK_H__

#include "mem.h"
#include "atomic.h"

typedef struct __mcspnode
{
	volatile int	            blocked;
	volatile struct __mcspnode* next;	
} mcspnode_t;

typedef struct
{
	volatile mcspnode_t*	tail;
} mcsplock_t;

#define DEFINE_MCSPLOCK(name) mcsplock_t name = { NULL }
#define DEFINE_MCSPNODE(name) mcspnode_t name


static inline void mcsp_lock(mcsplock_t* lock, mcspnode_t *self)
{
	mcspnode_t* prev;

	self->next     = NULL;

	do {
		prev = (mcspnode_t*) lock->tail;
	} while ((mcspnode_t*) cmpxchgptr(&lock->tail, prev, self)
		 != prev);
	
	if (prev) {
		self->blocked = 1;
		prev->next    = self;
		while (self->blocked) {
			cpu_relax();
		}
	}
}

static inline void mcsp_unlock(mcsplock_t* lock, mcspnode_t *self)
{
	if (!self->next && ((mcspnode_t*)
			    cmpxchgptr(&lock->tail, self, 0)) != self)
		while (!self->next)
			cpu_relax();
	if (self->next)
		self->next->blocked = 0;
}


#endif

