#ifndef MCSF_RW_PROTO_QLOCK
#define MCSF_RW_PROTO_QLOCK

#include "mem.h"
#include "atomic.h"
#include "locks/mcsrw_common.h"

#define RINC 0x2
#define PRES 0x1

static inline u32 mcsf_reader_count(mcs_rwlock_t* lock)
{
	return lock->count >> 1;
}

static inline void mcsf_write_lock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	mcs_rwnode_t* prev;

	*self = (mcs_rwnode_t) {.cls = WRITER, .next = NULL,
				.u.state.blocked = 1, .u.state.next_cls = NONE};

	prev = xchgptr(self, (void**)  &lock->tail);
	mb(); /* needed i386 */
	if (!prev) {
		lock->nextw = self;
		mb(); /* Can't have the compiler nor CPU reorder the store of
		       * next before the load of count, otherwise we could
		       * deadlock (as seen on both the Niagara by courtesy of
		       * the GCC optimizer and on Intel Xeon CPUs).
		       */
		if (xadd32(PRES, &lock->count) == PRES)
			self->u.state.blocked = 0;
	} else {
		prev->u.state.next_cls = WRITER;
		mb(); /* needed */
		prev->next = self;
	}

	while (self->u.state.blocked) {
		cpu_relax();
	}
}

static inline void mcsf_write_unlock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	mb();

	lock->count = 0; /* zero PRES bit */
	if (self->next ||
	    (void*) cmpxchgptr(&lock->tail, self, NULL) != self) {
		while (!self->next)
			cpu_relax();
		mb(); /* probably needed */
		if (self->next->cls == READER) {
			lock->count = RINC;
			mb();
		}
		self->next->u.state.blocked = 0;
	}
}

static inline void mcsf_read_lock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	mcs_rwnode_t* prev;
	 __mcs_atomic_u old, new;

	*self = (mcs_rwnode_t) {.cls = READER, .next = NULL,
				.u.state.blocked = 1, .u.state.next_cls = NONE};

	prev = xchgptr(self, (void**)  &lock->tail);
	if (!prev) {
		xadd32(RINC, &lock->count);
		self->u.state.blocked = 0;
	} else {
		old.state = (struct __mcs_rw_state) {1, NONE};
		new.state = (struct __mcs_rw_state) {1, READER};
		if (prev->cls == WRITER ||
		    cmpxchg32(&prev->u.raw, old.raw, new.raw) == old.raw) {
			prev->next = self;
			while (self->u.state.blocked)
				cpu_relax();
		} else {
			xadd32(RINC, &lock->count);
			prev->next = self;
			self->u.state.blocked = 0;
		}
	}
	mb(); /* needed  */
	if (self->u.state.next_cls == READER) {
		while (!self->next)
			cpu_relax();
		xadd32(RINC, &lock->count);
		self->next->u.state.blocked = 0;
	}
}

static inline void mcsf_read_unlock(mcs_rwlock_t* lock, mcs_rwnode_t *self)
{
	mb();

	if (self->next ||
	    ((void*) cmpxchgptr(&lock->tail, self, NULL)) != self) {
		mb();
		while (!self->next)
			cpu_relax();
//		mb(); /* needed? */
		if (self->u.state.next_cls == WRITER) {
			lock->nextw = (mcs_rwnode_t*) self->next;
			xadd32(PRES, &lock->count);
		}
	}
	if (xadd32(-RINC, &lock->count) == PRES)
		lock->nextw->u.state.blocked = 0;
}



#endif
