#ifndef MCSS_RW_PROTO_LOCK
#define MCSS_RW_PROTO_LOCK

#include "mem.h"
#include "atomic.h"

typedef struct {
	volatile u32 requests;
	volatile u32 completions;
} mcssp_rwlock_t;

#define RC_INCR   0x10000
#define WC_INCR   0x1
#define W_MASK    0xffff

#define WC_TOPMSK 0x8000
#define RC_TOPMSK 0x80000000

#define DEFINE_MCSSP_RWLOCK(x) mcssp_rwlock_t x =	\
	{ 0, 0 }

static inline u32 fetch_clear_then_add(volatile u32 *word, u32 mask, u32 increment)
{
	u32 old, new;

	do {
		old = *word;
		new = (old & ~mask) + increment;
	} while (cmpxchg32(word, old, new) != old);

	return old;
}

static inline void mcssp_read_lock(mcssp_rwlock_t *l)
{
	u32 prev_writers;

	prev_writers = fetch_clear_then_add(&l->requests, RC_TOPMSK, RC_INCR);
	prev_writers &= W_MASK;

	while ((l->completions & W_MASK) != prev_writers)
		cpu_relax();
}

static inline void mcssp_read_unlock(mcssp_rwlock_t *l)
{
	fetch_clear_then_add(&l->completions, RC_TOPMSK, RC_INCR);
}


static inline void mcssp_write_lock(mcssp_rwlock_t *l)
{
	u32 prev =  fetch_clear_then_add(&l->requests, WC_TOPMSK, WC_INCR);

	while (l->completions != prev)
		cpu_relax();
}

static inline void mcssp_write_unlock(mcssp_rwlock_t *l)
{
	fetch_clear_then_add(&l->completions, WC_TOPMSK, WC_INCR);
}

#endif
