#ifndef RWLOCKS_H
#define RWLOCKS_H

#include "kernel_compat.h"
#include "mem.h"
#include "atomic.h"

#if defined(USE_LIN_RWLOCK)

#include "locks/linrw.h"
#define DEF_RW_LOCK(x) DEFINE_LIN_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x)
#define read_lock(x, ctx)      lin_read_lock(x)
#define read_unlock(x, ctx)    lin_read_unlock(x)
#define write_lock(x, ctx)     lin_write_lock(x)
#define write_unlock(x, ctx)   lin_write_unlock(x)

#elif defined(USE_MCS_PROTO_RWLOCK)

#include "locks/mcsrw_proto.h"
#define DEF_RW_LOCK(x) DEFINE_MCS_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x) DEFINE_MCS_RWNODE(x)
#define HAVE_RW_LOCK_CTX
#define reader_count(x)        mcs_reader_count(x)
#define HAVE_RW_READER_COUNT
#define read_lock(x, ctx)      mcsp_read_lock(x, &ctx)
#define read_unlock(x, ctx)    mcsp_read_unlock(x, &ctx)
#define write_lock(x, ctx)     mcsp_write_lock(x, &ctx)
#define write_unlock(x, ctx)   mcsp_write_unlock(x, &ctx)

#elif defined(USE_MCSF_PROTO_RWLOCK)

#include "locks/mcsrwf_proto.h"
#define DEF_RW_LOCK(x) DEFINE_MCS_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x) DEFINE_MCS_RWNODE(x)
#define HAVE_RW_LOCK_CTX
#define reader_count(x)        mcsf_reader_count(x)
#define HAVE_RW_READER_COUNT
#define read_lock(x, ctx)      mcsf_read_lock(x, &ctx)
#define read_unlock(x, ctx)    mcsf_read_unlock(x, &ctx)
#define write_lock(x, ctx)     mcsf_write_lock(x, &ctx)
#define write_unlock(x, ctx)   mcsf_write_unlock(x, &ctx)

#elif defined(USE_MCS_RWLOCK)

#include "locks/mcsrw.h"
#define DEF_RW_LOCK(x) DEFINE_MCS_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x) DEFINE_MCS_RWNODE(x)
#define HAVE_RW_LOCK_CTX
#define reader_count(x)        mcs_reader_count(x)
#define HAVE_RW_READER_COUNT
#define read_lock(x, ctx)      mcs_read_lock(x, &ctx)
#define read_unlock(x, ctx)    mcs_read_unlock(x, &ctx)
#define write_lock(x, ctx)     mcs_write_lock(x, &ctx)
#define write_unlock(x, ctx)   mcs_write_unlock(x, &ctx)

#elif defined(USE_BA_PROTO1_RWLOCK)

#include "locks/barw_proto1.h"
#define DEF_RW_LOCK(x) DEFINE_BA_PROTO1_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x)
#define read_lock(x, ctx)      ba_proto1_read_lock(x)
#define read_unlock(x, ctx)    ba_proto1_read_unlock(x)
#define write_lock(x, ctx)     ba_proto1_write_lock(x)
#define write_unlock(x, ctx)   ba_proto1_write_unlock(x)

#elif defined(USE_BA_PROTO2_RWLOCK)

#include "locks/barw_proto2.h"
#define DEF_RW_LOCK(x) DEFINE_BA_PROTO2_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x)
#define read_lock(x, ctx)      ba_proto2_read_lock(x)
#define read_unlock(x, ctx)    ba_proto2_read_unlock(x)
#define write_lock(x, ctx)     ba_proto2_write_lock(x)
#define write_unlock(x, ctx)   ba_proto2_write_unlock(x)

#elif defined(USE_PF_RWLOCK)

#include "locks/pfrw.h"
#define DEF_RW_LOCK(x) DEFINE_PF_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x)
#define read_lock(x, ctx)      pf_read_lock(x)
#define read_unlock(x, ctx)    pf_read_unlock(x)
#define write_lock(x, ctx)     pf_write_lock(x)
#define write_unlock(x, ctx)   pf_write_unlock(x)

#elif defined(USE_MCSS_RWLOCK)

#include "locks/mcssrw.h"
#define DEF_RW_LOCK(x) DEFINE_MCSS_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x)
#define read_lock(x, ctx)      mcss_read_lock(x)
#define read_unlock(x, ctx)    mcss_read_unlock(x)
#define write_lock(x, ctx)     mcss_write_lock(x)
#define write_unlock(x, ctx)   mcss_write_unlock(x)

#elif defined(USE_MCSS_PROTO_RWLOCK)

#include "locks/mcssrw_proto.h"
#define DEF_RW_LOCK(x) DEFINE_MCSSP_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x)
#define read_lock(x, ctx)      mcssp_read_lock(x)
#define read_unlock(x, ctx)    mcssp_read_unlock(x)
#define write_lock(x, ctx)     mcssp_write_lock(x)
#define write_unlock(x, ctx)   mcssp_write_unlock(x)

#elif defined(USE_PFQ_PROTO_RWLOCK)

#include "locks/pfq_proto.h"
#define DEF_RW_LOCK(x) DEFINE_PFQ_RWLOCK(x)
#define DEF_RW_LOCK_CTX(x) DEFINE_PFQ_RWNODE(x)
#define HAVE_RW_LOCK_CTX
#define read_lock(x, ctx)      pfqp_read_lock(x, &ctx)
#define read_unlock(x, ctx)    pfqp_read_unlock(x, &ctx)
#define write_lock(x, ctx)     pfqp_write_lock(x, &ctx)
#define write_unlock(x, ctx)   pfqp_write_unlock(x, &ctx)

#elif defined(USE_MUTEX_EMULATION_RWLOCK)

#include "locks/locks.h"
#define DEF_RW_LOCK(x) DEF_LOCK(x)
#define DEF_RW_LOCK_CTX(x) DEF_LOCK_CTX(x)
#ifdef HAVE_LOCK_CTX
#define HAVE_RW_LOCK_CTX
#endif
#define read_lock(x, ctx)      lock(x, ctx)
#define read_unlock(x, ctx)    unlock(x, ctx)
#define write_lock(x, ctx)     lock(x, ctx)
#define write_unlock(x, ctx)   unlock(x, ctx)

#elif defined(USE_NONE_RWLOCK)

#define DEF_RW_LOCK(x)
#define DEF_RW_LOCK_CTX(x)
#define read_lock(x, ctx)
#define read_unlock(x, ctx)
#define write_lock(x, ctx)
#define write_unlock(x, ctx)

#else

#error "What kind of RW lock should I use?"

#endif


#endif
