#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <pthread.h>

#include <linux/unistd.h>
#include <sys/types.h>
#include <sched.h>

#include "parallel.h"

#include "kernel_compat.h"
#include "mem.h"
#include "atomic.h"
#include "cycles.h"

#include "locks/rwlocks.h"

#define WARMUP    100
//#define LOOPS 3000000
#define LOOPS   1000000

cycles_t rec_enter[LOOPS];
cycles_t rec_exit[LOOPS];

void bench_read(void)
{
	int i;
	DEF_RW_LOCK(s);
	DEF_RW_LOCK_CTX(x);
	for (i = 0; i < WARMUP; i++) {
		read_lock(&s, x);
		cpu_relax();
		read_unlock(&s, x);
	}

	for (i = 0; i < LOOPS; i++) {
		rec_enter[i] = get_cycles();
		read_lock(&s, x);
		cpu_relax();
		read_unlock(&s, x);
		rec_exit[i] = get_cycles();
	}
}

void bench_write(void)
{
	int i;
	DEF_RW_LOCK(s);
	DEF_RW_LOCK_CTX(x);
	for (i = 0; i < WARMUP; i++) {
		write_lock(&s, x);
		cpu_relax();
		write_unlock(&s, x);
	}

	for (i = 0; i < LOOPS; i++) {
		rec_enter[i] = get_cycles();
		write_lock(&s, x);
		cpu_relax();
		write_unlock(&s, x);
		rec_exit[i] = get_cycles();
	}
}

void write_csv(char* fname)
{
	int i;
	FILE *f;
	f = fopen(fname, "w");
	for (i = 0; i < LOOPS; i++)
		fprintf(f, "%" CYCLES_FMT ", %" CYCLES_FMT 
			", %" CYCLES_FMT "\n",
			rec_enter[i], rec_exit[i], rec_exit[i] - rec_enter[i]);
	fclose(f);
}

int main(int argc, char** argv)
{
	char* basename;
	char rfile[120], wfile[120];
	if (argc != 2) {
		printf("usage: %s <basename>\n", argv[0]);
		exit(1);
	}
	basename = argv[1];
	sprintf(rfile, "%s-reads.csv", basename);
	sprintf(wfile, "%s-writes.csv", basename);
	printf("Benchmarking reads...\n");
	bench_read();
	printf("Writing to %s...\n", rfile);
	write_csv(rfile);
	printf("Benchmarking writes...\n");
	bench_write();
	printf("Writing to %s...\n", wfile);
	write_csv(wfile);
	return 0;
}
