#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <string.h>

#include "units.h"
#include "parallel.h"
#include "cycles.h"

#define KB 1024
#define MB 1024 * KB
#define MAX_CPUS 32

int disable_migration = 0;
static void migrate(void)
{
	int cpu = rand() % MAX_CPUS;
	if (!disable_migration)
		migrate_to(cpu);
}

static void error(const char* msg)
{
	fprintf(stderr, "Error: %s\n", msg);
	exit(1);
}

int stop = 0;

static void on_signal(int sig)
{
	if (!stop)
		stop = 1;
	else
		error("Double signal, exiting.");
}

static cycles_t iterate(int *mem, int length, int chunk)
{
	int x, y, z, i;
	cycles_t t0, t1;
	t0 = get_cycles();
	for (i = 0; i < chunk; i++) {
		x = rand() % length;
		y = rand() % length;
		z = rand() % length;
		mem[z] = mem[x] + mem[y] - rand();
	}
	t1 = get_cycles();
	return t1 - t0;
}

static void run(int wss, int chunk)
{
        cycles_t total_time = 0, total_delta = 0, warm, cold;
	unsigned long chunks = 0;
	int length = wss / sizeof(int);
	int* mem;
	double cost, delta_cost;
	mem = malloc(length * sizeof(int));
	
	if (!mem)
		error("malloc() failed");

	while (!stop) {
		migrate();
		cold = iterate(mem, length, chunk / 2);
		warm = iterate(mem, length, chunk / 2);
		total_time += cold + warm;
		if (cold > warm)
			total_delta += (cold - warm);
		chunks++;
	}
	cost = total_time / (chunks * (double) chunk);
	delta_cost = total_delta / (double) chunks;
	printf("%10d, %10d, %10lu, %10.2f, %10.2f\n",
	       wss, chunk, chunks, cost, delta_cost);
	fflush(stdout);
}


#define OPTSTR "w:c:t:d"

int main(int argc, char** argv)
{
	int opt;
	int wss  = 3 * MB;
	int chunk = 100000;
	int timeout = 0;
	float tmp;

	srand(time(NULL));

	signal(SIGINT, on_signal);
	signal(SIGTERM, on_signal);
	signal(SIGUSR1, on_signal);
	signal(SIGALRM, on_signal);

	migrate_to(0);

	while ((opt = getopt(argc, argv, OPTSTR)) != -1) {
		switch (opt) {
		case 'w':
			if (!str2float(optarg, &tmp))
				error("malformed argument");
			wss = tmp;
			if (wss <= 0)
				error("wss <= 0");
			break;
		case 'c':
			if (!str2float(optarg, &tmp))
				error("malformed argument");
			chunk = tmp;
			if (chunk <= 0)
				error("chunk <= 0");
			break;
		case 't':
			if (!str2int(optarg, &timeout))
				error("malformed argument");
			if (timeout < 0)
				error("timeout < 0");
			break;
		case 'd':
			disable_migration = 1;
			break;
		case ':':
			error("Argument missing.");
			break;
		case '?':
		default:
			error("Bad argument.");
			break;
		}
	}
	
	alarm(timeout);
	run(wss, chunk);

	return 0;
}
