#include <stdio.h>
#include <stdlib.h>

#include <pthread.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <sys/types.h>
#include <sched.h>

#include "parallel.h"

struct worker_info {
	int			id;
	void*			arg;
	worker_t		func;
	pthread_barrier_t*	launch_barrier;
	int			set_affinity;
};

pid_t gettid(void)
{
	return syscall(__NR_gettid);
}

int migrate_to(unsigned int cpu)
{
	cpu_set_t mask;
	CPU_ZERO(&mask);
	CPU_SET(cpu, &mask);
	return sched_setaffinity(gettid(), sizeof(mask), &mask);
}

static void* __startup(void* _winfo) 
{
	struct worker_info* winfo = (struct worker_info*) _winfo;

	if (winfo->set_affinity && migrate_to(winfo->id) != 0)
		perror("migrate_to() failed");

	if (winfo->launch_barrier) {
		pthread_barrier_wait(winfo->launch_barrier);
	}

	winfo->func(winfo->id, winfo->arg);
	return NULL;
}

void in_parallel(unsigned int num_threads, worker_t worker, void* arg,
		 int use_barrier, int set_affinity)
{
	unsigned int i;
	pthread_t *threads = malloc(sizeof(pthread_t) * num_threads);
	pthread_barrier_t launch_barrier;
	struct worker_info* info = 
		malloc(sizeof(struct worker_info) * num_threads);

	if (use_barrier)
		pthread_barrier_init(&launch_barrier, NULL, num_threads);

	for (i = 0; i < num_threads; i++) {
		info[i].id   = i;
		info[i].arg  = arg;
		info[i].func = worker;
		info[i].launch_barrier = use_barrier ?  &launch_barrier : NULL;
		info[i].set_affinity   = set_affinity;
	}

	for (i = 0; i < num_threads; i++)
		if (pthread_create(threads + i, NULL, __startup, info + i) != 0)
			perror("pthread_create");		
	
	
	for (i = 0; i < num_threads; i++)
		pthread_join(threads[i], NULL);

	if (use_barrier)
		pthread_barrier_destroy(&launch_barrier);
}
