#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "parallel.h"
#include "locks/locks.h"
#include "cycles.h"

DEF_LOCK(global_lock);
int counter = 0;
int last_id = -1;
int repetition_counter = 0;

#define LOOPS_PER_THREAD 200000

int num_threads = 0;

cycles_t min[MAX_THREADS], max[MAX_THREADS], total[MAX_THREADS], exec[MAX_THREADS];
int local_rep[MAX_THREADS];

void init(void)
{
	int i;
	counter = 0;
	for (i = 0; i < num_threads; i++) {
		min[i] = ULONG_MAX;
		max[i] = 0;
		total[i] = 0;
		local_rep[i] = 0;
	}
}



void show_stats(void)
{
	int i;
	int max_blk, min_blk, max_total_blk, min_total_blk, max_exe, min_exe;
	max_blk = min_blk = max_total_blk = min_total_blk = max_exe = min_exe = 0;
	for (i = 0; i < num_threads; i++) {
		printf("[%3d] blocking: min=%7.3fus avg=%7.3fus max=%10.3fus rep=%7d\n"
		       "      exe=%10.3fms\n"
		       "      blk=%10.3fms\n"
		       "      ovh=%10.3fms\n", i,
		       cycles2usec(min[i]),
		       cycles2usec(total[i]) / (double) LOOPS_PER_THREAD,
		       cycles2usec(max[i]),
		       local_rep[i],
		       cycles2usec(exec[i]) / 1000,
		       cycles2usec(total[i]) / 1000,
		       cycles2usec(exec[i] - total[i]) / 1000);
		if (max[i] > max[max_blk])
			max_blk = i;
		if (max[i] < max[min_blk])
			min_blk = i;
		if (total[i] > total[max_total_blk])
			max_total_blk = i;
		if (total[i] < total[min_total_blk])
			min_total_blk = i;
		if (exec[i] < exec[min_exe])
			min_exe = i;
		if (exec[i] > exec[max_exe])
			max_exe = i;
	}
	printf("all: min(max)=%10.3fus (%3d)  max(max)=%10.3fus (%3d)\n"
	       "     delta=%10.3fus (per thread)\n"
	       "     min(blk)=%10.3fms (%3d)  max(blk)=%10.3fms (%3d)\n"
	       "     delta=%10.3fms (per thread)\n"
	       "     min(exe)=%10.3fms (%3d)  max(exe)=%10.3fms (%3d)\n"
	       "     delta=%10.3fms (per thread)\n",
	       cycles2usec(max[min_blk]), min_blk,
	       cycles2usec(max[max_blk]), max_blk,
	       cycles2usec(max[max_blk] -  max[min_blk]) / num_threads,
	       cycles2usec(total[min_total_blk]) / 1000, min_total_blk,
	       cycles2usec(total[max_total_blk]) / 1000, max_total_blk,
	       cycles2usec((total[max_total_blk] - total[min_total_blk])
			   / 1000) / num_threads,
	       cycles2usec(exec[min_exe]) / 1000, min_exe,
	       cycles2usec(exec[max_exe]) / 1000, max_exe,
	       cycles2usec((exec[max_exe] - exec[min_exe])
			   / 1000) / num_threads);

}

void show_csv(void)
{
	int i;
	int max_blk, min_blk, max_total_blk, min_total_blk, max_exe, min_exe;
	double sum_exe = 0;
	max_blk = min_blk = max_total_blk = min_total_blk = max_exe = min_exe = 0;
	for (i = 0; i < num_threads; i++) {
		if (max[i] > max[max_blk])
			max_blk = i;
		if (max[i] < max[min_blk])
			min_blk = i;
		if (total[i] > total[max_total_blk])
			max_total_blk = i;
		if (total[i] < total[min_total_blk])
			min_total_blk = i;
		if (exec[i] < exec[min_exe])
			min_exe = i;
		if (exec[i] > exec[max_exe])
			max_exe = i;
		sum_exe += cycles2usec(exec[i]);
	}
	printf("%3d, %10.4f, %10.4f, %10.4f, %10.4f, %10.4f, %10.4f\n",
	       num_threads,
	       cycles2usec(max[max_blk] -  max[min_blk]) / num_threads,
	       cycles2usec((total[max_total_blk] - total[min_total_blk])
			   / 1000) / num_threads,
	       cycles2usec((exec[max_exe] - exec[min_exe])
			   / 1000) / num_threads,
	       cycles2usec(exec[max_exe]) / counter,
	       cycles2usec(exec[max_exe]) / 1000,
	       sum_exe / num_threads);
}

void bang_on_lock(int id, void* _)
{
	int i, tmp;
	cycles_t start, stop, t1, t2;
	DEF_LOCK_CTX(lock_ctx);

	t1 = get_cycles();
	for (i = 0; i < LOOPS_PER_THREAD; i++) {
		start = get_cycles();
		lock(&global_lock, lock_ctx);
		stop  = get_cycles();
		stop -= start;
		if (stop < min[id])
			min[id] = stop;
		if (stop > max[id])
			max[id] = stop;
		total[id] += stop;
		tmp = counter;
		counter++;
		repetition_counter += last_id == id;
		local_rep[id] += last_id == id;
		last_id = id;
		unlock(&global_lock, lock_ctx);
	}
	t2 = get_cycles();
	exec[id] = t2 - t1;
}

int main(int argc, char** argv)
{
	int warmup =  3;
	int sample = 2;
	int j;

	warmup = 1;
//	sample = 3;
	for (num_threads = 2; num_threads <= MAX_THREADS; num_threads++) {
		for (j = 0; j < warmup; j++) {
			init();
			in_parallel(num_threads, bang_on_lock, NULL, 1, 1);
			if (counter != LOOPS_PER_THREAD * num_threads)
				printf("counter = %d (expected %d)\n", counter,
				       LOOPS_PER_THREAD * num_threads);
		}
		for (j = 0; j < sample; j++) {
			init();
			in_parallel(num_threads, bang_on_lock, NULL, 1, 1);
			if (counter != LOOPS_PER_THREAD * num_threads)
				printf("counter = %d (expected %d)\n", counter,
				       LOOPS_PER_THREAD * num_threads);
			show_csv();
			//show_stats();
		}

	}

	return 0;
}
